@@ -1,7 +1,7 @@
 """
 DATABRICKS_HOST, DATABRICKS_TOKEN
 """
-
+import collections
 import json
 import os
 import re
@@ -55,6 +55,7 @@ class ProphecyBuildTool:
 
     def build(self):
         print('\n[bold blue]Building %s pipelines 🚰[/bold blue]' % self.pipelines_count)
+        overall_build_status = True
         for pipeline_i, (path_pipeline, pipeline) in enumerate(self.pipelines.items()):
             print('\n  Building pipeline %s [%s/%s]' % (path_pipeline, pipeline_i + 1, self.pipelines_count))
 
@@ -93,41 +94,63 @@ class ProphecyBuildTool:
                 if build_file_found:
                     print("\n[bold blue]✅ Build complete![/bold blue]")
                 else:
-                    print("\n[bold red]❌ Build completed but built target not found![/bold red]")
-                    sys.exit(1)
+                    print(
+                        f"\n[bold red]❌ Build completed but built target not found for pipeline: {pipeline['name']}![/bold red]")
+                    overall_build_status = False
             else:
-                print("\n[bold red]❌ Build failed![/bold red]")
-                sys.exit(1)
+                print(f"\n[bold red]❌ Build failed for pipeline: {pipeline['name']}![/bold red]")
+                overall_build_status = False
+
+        if not overall_build_status:
+            sys.exit(1)
 
     def deploy(self):
         self.build()
 
         print('\n[bold blue]Deploying %s jobs ⏱[/bold blue]' % self.jobs_count)
+
+        pipelines_upload_failures = collections.defaultdict(list)
+        job_update_failures = dict()
+
         for job_idx, (path_job, job) in enumerate(self.jobs.items()):
             print('\n  Deploying job %s [%s/%s]' % (path_job, job_idx + 1, self.jobs_count))
 
             path_job_absolute = os.path.join(os.path.join(self.path_root, path_job), 'code')
             path_job_definition = os.path.join(path_job_absolute, 'databricks-job.json')
 
-            job_definition = {}
             with open(path_job_definition, 'r') as _in:
                 job_definition = json.load(_in)
 
             components = job_definition['components']
+
             for component in components:
                 if 'PipelineComponent' in component:
                     pipeline_uri = component['PipelineComponent']['id']
 
-                    uri_pattern = '([0-9]+)/([-_.A-Za-z0-9 /]+)'
+                    # Matches project_id/pipelines/pipeline_name or pipelines/pipeline_name or /pipelines/pipeline_name
+                    # group(2) should return pipelines/pipeline_name
+                    uri_pattern = '([0-9]*)(pipelines/[-_.A-Za-z0-9 /]+)'
                     pipeline_id = re.search(uri_pattern, pipeline_uri).group(2)
 
                     source_path = self.pipelines_build_path[pipeline_id]['source_absolute']
                     target_path = component['PipelineComponent']['path']
 
                     print('    Uploading %s to %s' % (self.pipelines_build_path[pipeline_id]['source'], target_path))
-                    self.dbfs_service.put(target_path, overwrite=True, src_path=source_path)
 
-                    self.pipelines_build_path[pipeline_id]['uploaded'] = True
+                    try:
+                        self.dbfs_service.put(target_path, overwrite=True, src_path=source_path)
+                        self.pipelines_build_path[pipeline_id]['uploaded'] = True
+                    except Exception as e:
+                        self.pipelines_build_path[pipeline_id]['uploaded'] = False
+                        pipelines_upload_failures[pipeline_id].append(str(e))
+
+            if len(pipelines_upload_failures) > 0:
+                print("\n[bold red]❌ Upload failed for one or more pipelines [/bold red]")
+                for pipeline_id, errors in pipelines_upload_failures.items():
+                    print('\n[bold red]❌ %s: Exceptions: %s [/bold red]' % (pipeline_id, "\n".join(errors)))
+
+                # Process next job
+                continue
 
             job_request = job_definition['request']['CreateNewJobRequest']
             job_request['version'] = '2.1'
@@ -135,33 +158,42 @@ class ProphecyBuildTool:
             limit = 25
             current_offset = 0
             found_job = None
-            while found_job is None:
-                response = self.jobs_service.list_jobs(limit=limit, offset=current_offset, version='2.1')
-                current_offset += limit
+            try:
+                while found_job is None:
+                    response = self.jobs_service.list_jobs(limit=limit, offset=current_offset, version='2.1')
+                    current_offset += limit
 
-                found_jobs = response['jobs'] if 'jobs' in response else []
-                for potential_found_job in found_jobs:
-                    if potential_found_job['settings']['name'] == job_request['name']:
-                        found_job = potential_found_job
+                    found_jobs = response['jobs'] if 'jobs' in response else []
+                    for potential_found_job in found_jobs:
+                        if potential_found_job['settings']['name'] == job_request['name']:
+                            found_job = potential_found_job
+                            break
+
+                    if found_job is not None or len(found_jobs) <= 0:
                         break
 
-                if found_job is not None or len(found_jobs) <= 0:
-                    break
+                job_request = job_definition['request']['CreateNewJobRequest']
+                if found_job is None:
+                    print('    Creating a new job: %s' % (job_request['name']))
 
-            job_request = job_definition['request']['CreateNewJobRequest']
-            if found_job is None:
-                print('    Creating a new job: %s' % (job_request['name']))
-                keys_delete = ['format', 'job_clusters']
+                    self.jobs_service.create_job(**job_request)
+                else:
+                    print('    Updating an existing job: %s' % (job_request['name']))
+                    self.jobs_service.reset_job(found_job['job_id'], new_settings=job_request, version='2.1')
+            except Exception as e:
+                print(
+                    f"\n[bold red]❌ Create/Update for job: {job_request['name']} failed with exception: {e} [/bold red]")
+                job_update_failures[job_request['name']] = str(e)
 
-                for key in keys_delete:
-                    if key in job_request:
-                        del job_request[key]
-                self.jobs_service.create_job(**job_request)
-            else:
-                print('    Updating an existing job: %s' % (job_request['name']))
-                self.jobs_service.reset_job(found_job['job_id'], new_settings=job_request, version='2.1')
-
-        print("\n[bold blue]✅ Deployment complete![/bold blue]")
+        if len(pipelines_upload_failures) == 0 and len(job_update_failures) == 0:
+            print("\n[bold blue]✅ Deployment completed successfully![/bold blue]")
+        else:
+            print("\n[bold red]❌ Deployment failed![/bold red]")
+            if len(pipelines_upload_failures) > 0:
+                print("   Upload failed for pipelines: %s" % (" ,".join(pipelines_upload_failures.keys())))
+            if len(job_update_failures) > 0:
+                print("   Create/Update failed for jobs: %s" % (" ,".join(job_update_failures.keys())))
+            sys.exit(1)
 
     def run_unit_tests(self):
         if self._verify_unit_test_env():
@@ -195,9 +227,9 @@ class ProphecyBuildTool:
             [
                 # Install dependencies of particular pipeline
                 # 1. Export dependencies to egg_info/requires.txt
-                Process(['python', 'setup.py', 'egg_info'], path_pipeline_absolute),
+                Process(['python', 'setup.py', '-q', 'egg_info'], path_pipeline_absolute),
                 # 2. Extract the install_requires, tests_requires and extra_requires and install them
-                Process([f"pip install `grep -v '^\[' *.egg-info/requires.txt`"],
+                Process([f"pip install -q `grep -v '^\[' *.egg-info/requires.txt`"],
                         path_pipeline_absolute, is_shell=True),
                 # Check for compilation errors
                 Process(['python', '-m', 'compileall', '.', '-q'], path_pipeline_absolute),
@@ -213,7 +245,7 @@ class ProphecyBuildTool:
     def test_scala(self, path_pipeline_absolute):
         return self._process_sequential(
             [
-                Process(['mvn', 'test', '-q', '-Dfabric=' + self.fabric, 'surefire-report:report'],
+                Process(['mvn', 'test', '-q', '-Dfabric=' + self.fabric],
                         path_pipeline_absolute)
             ]
         )
@@ -223,9 +255,9 @@ class ProphecyBuildTool:
             [
                 # Install dependencies of particular pipeline
                 # 1. Export dependencies to egg_info/requires.txt
-                Process(['python', 'setup.py', 'egg_info'], path_pipeline_absolute),
+                Process(['python', 'setup.py', '-q', 'egg_info'], path_pipeline_absolute),
                 # 2. Extract the install_requires, tests_requires and extra_requires and install them
-                Process([f"pip install `grep -v '^\[' *.egg-info/requires.txt`"],
+                Process([f"pip install -q `grep -v '^\[' *.egg-info/requires.txt`"],
                         path_pipeline_absolute, is_shell=True),
                 # Run the unit test
                 Process(['python', '-m', 'pytest', '-v', 'test/TestSuite.py', '--html=report.html',
@@ -263,9 +295,9 @@ class ProphecyBuildTool:
             map(lambda pipeline: '%s (%s)' % (pipeline['name'], pipeline['language']), self.pipelines.values()))
         print('Found %s pipelines: %s' % (self.pipelines_count, pipelines_str))
 
-    def _process_sequential(self, subprocesses, time_between_each_cmd=1):
+    def _process_sequential(self, processes, time_between_each_cmd=1):
         return_code = 0
-        for process in subprocesses:
+        for process in processes:
             result = subprocess.run(process.process_args, stdout=process.std_output, stderr=process.std_err,
                                     shell=process.is_shell, cwd=process.current_working_directory)
             return_code, stdout, stderr = result.returncode, result.stdout, result.stderr
@@ -276,7 +308,6 @@ class ProphecyBuildTool:
             else:
                 print('   ', '\n    '.join([line.decode('utf-8') for line in stdout.splitlines()]))
 
-            print("Return code " + str(return_code))
             if return_code != 0:
                 break
             time.sleep(time_between_each_cmd)
